//---------------------------------------------------------------------------
// Written 1995 By Dan Bystrm
//---------------------------------------------------------------------------
// DBTTip.c
//---------------------------------------------------------------------------

#include "resource.h"
#include "dbttip.h"

#include <stdio.h>

//---------------------------------------------------------------------------
// Global Variables
//---------------------------------------------------------------------------
HANDLE ghmodDLL;
HANDLE taskDevEnvironment = NULL;

BOOL RegisterTipClass(void);
void UnregisterTipClass(void);

void CreateTipWnd(HCTL);
void DestroyTipWnd(HCTL);

//---------------------------------------------------------------------------
// Standard Error Values
//---------------------------------------------------------------------------
#define ERR_None	      0
#define ERR_InvPropVal	    380     // Error$(380) = "Invalid property value"


//---------------------------------------------------------------------------
// Event Procedure Parameter Profiles
//---------------------------------------------------------------------------

//New function for v1.2
BOOL GoDown( HCTL hctl )
{
 LPDBTTIP lpDBTTip;

 DestroyTipWnd( hctl );
 lpDBTTip = LPDBTTIPDEREF(hctl);
 lpDBTTip->nCountDown = lpDBTTip->nDelay / 50;
 if ( lpDBTTip->TipStatus != TIPPING_NONE )
  {
   lpDBTTip->TipStatus = TIPPING_NONE ;
   return VBFireEvent( hctl, IEVENT_TTIP_POPDOWN, NULL );
  }
 return FALSE;
}


//New function for v1.2
// hctl1 will be verified
// hctl2 must be an existing container
BOOL VerifyHCTL2( HCTL hctl1, HCTL hctl2 )
{
 if ( hctl1 == hctl2 )
   return TRUE;
 hctl2 = VBGetControl( hctl2, GC_CHILD );
 while ( hctl2 )
  {
   if ( VerifyHCTL2( hctl1, hctl2 ) )
     return TRUE;
   hctl2 = VBGetControl( hctl2, GC_NEXTSIBLING );
  }
 return FALSE;
}


//New function for v1.2
// hctl will be verified
// hwnd must be hwnd of a known container (or the form or the control itself)
BOOL VerifyHCTL( HCTL hctl, HWND hwnd )
{
 if ( !IsWindow(hwnd) )
   return FALSE;

 return VerifyHCTL2( hctl, VBGetHwndControl(hwnd) );
}


//New function for v1.23
// copies a string an substitutes all || with CRLF
void PipeCopy( LPSTR lpszD, LPCSTR lpszS )
{
 LPSTR ptr;

 lstrcpy( lpszD, lpszS );
 for ( ptr=lpszD ; *ptr ; ptr++ )
   if ( ptr[0]=='|' && ptr[1]=='|' )
    {
     ptr[0] = 13;
     ptr[1] = 10;
    }
}


void AdjustTipSize( HCTL hctl )
{
 HWND hWnd = VBGetControlHwnd(hctl);
 HDC hDC = GetDC( hWnd );
 LPDBTTIP lpDBTTip;
 DWORD dwWH;
 HFONT hFontOld = NULL;
 RECT rect;
 short nLineHeight, nLineWidth, nBottom, nRight, nRightTest;
 char szText[1200];

 PipeCopy( szText, VBDerefHsz(LPDBTTIPDEREF(hctl)->hszText) );
 lpDBTTip = LPDBTTIPDEREF(hctl);

 //Select the correct font and get the dimension of a one line caption
 if ( lpDBTTip->uUse.hFont )
   hFontOld = SelectObject( hDC, lpDBTTip->uUse.hFont );
 dwWH = GetTextExtent( hDC, szText, lstrlen(szText) );
 nLineHeight = HIWORD(dwWH);
 nLineWidth = LOWORD(dwWH);

 //Make a bounding box of the caption
 rect.left	= 0;
 rect.right	= nLineWidth;
 rect.top	= 0;
 rect.bottom= nLineHeight;
 nBottom	= nLineHeight;

 //While the aspect doesn't match (= the box is too wide)
 //increase the caption by one line and calculate a new bounding box
 if ( (long)rect.right*(long)lpDBTTip->uUse.nAspect > (long)rect.bottom )
  {
   do {
     nBottom += nLineHeight;
     rect.right = (nBottom * lpDBTTip->uUse.nAspect) / 10;
     rect.bottom = DrawText(hDC, szText, -1, &rect, DT_CALCRECT|DT_WORDBREAK );
    } while ( rect.bottom > nBottom );
   // Narrow the box as much as possible without getting an extra line
   nBottom = rect.bottom;
   nRightTest = rect.right;
   do {
     nRight = rect.right;
     if ( (nRightTest -= 10) < 5 )
       break;
     rect.right = nRightTest;
    } while ( nBottom==DrawText(hDC, szText, -1, &rect, DT_CALCRECT|DT_WORDBREAK ) );
   rect.right  = nRight;
   rect.bottom = nBottom;
  }

 //Don't need the hDC anymore
 if ( hFontOld )
   SelectObject( hDC, hFontOld );
 ReleaseDC( hWnd, hDC );

 //950607 Add an extra 2 pixels on the left and right side of the box
 rect.right += 4;

 //Now the size of the bounding box is correct, but not it's offset
 //Store the box and fix the offset later
 CopyRect( &lpDBTTip->rectText, &rect );

 //Based on the bounding box' size, we are now able to calculate
 //the size of the tooltip window
 lpDBTTip->nTipWidth  = (short)( (LONG)rect.right*100L / (LONG)(lpDBTTip->uUse.nTextRight-lpDBTTip->uUse.nTextLeft) );
 lpDBTTip->nTipHeight = (short)( (LONG)rect.bottom*100L / (LONG)(lpDBTTip->uUse.nTextBottom-lpDBTTip->uUse.nTextTop) );

 //Now fix the offset of the caption's bounding box
 OffsetRect( &lpDBTTip->rectText,
             (short)( (LONG)lpDBTTip->nTipWidth*(LONG)lpDBTTip->uUse.nTextLeft / 100L) +2, //950607 Added +2
             (short)( (LONG)lpDBTTip->nTipHeight*(LONG)lpDBTTip->uUse.nTextTop / 100L ) );
}


#ifdef _DEBUG
#include "stdarg.h"
void CDECL TELL(LPCSTR pszFormat, ...)
{
  static char szBuffer[512];
  va_list args;
  
  va_start(args, pszFormat);
  wvsprintf( szBuffer, pszFormat, args );
  lstrcat( szBuffer, "\r\n" );
  OutputDebugString (szBuffer);
}
#else
#define TELL(_x_)
#endif

void PaintTip( HCTL hctl, HDC hDC, RECT rect, HMETAFILE hmf, int xExt, int yExt )
{
 LPDBTTIP lpDBTTip = LPDBTTIPDEREF(hctl);
 HBRUSH hOldBrush = SelectObject( hDC, GetStockObject(NULL_BRUSH) );
 HPEN hOldPen;
 short width, height, sdc;

 //Draw the black frame
 Rectangle( hDC, rect.left, rect.top, rect.right, rect.bottom );

 //Calculate bounding box for WMF
 rect.left	= rect.right/8;
 rect.top	= rect.bottom/8;
 rect.right	-= rect.left;
 rect.bottom-= rect.top;
 width = rect.right - rect.left;
 height = rect.bottom - rect.top;

 //Draw the WMF
 sdc = SaveDC( hDC );;
 SetMapMode( hDC, MM_ANISOTROPIC );
 SetWindowExt( hDC, xExt, yExt );
 SetViewportExt( hDC, width, height );
 SetViewportOrg( hDC, rect.left, rect.top );
 PlayMetaFile( hDC, hmf );
 RestoreDC( hDC, sdc );

 //Draw the bounding box of the WMF
 DrawFocusRect( hDC, &rect );

 { //A circle around the hot spot
  short nHotSpotX = rect.left + (short)( ((LONG)width*(LONG)lpDBTTip->uMe.nHotSpotX) / 100L  );
  short nHotSpotY = rect.top + (short)( ((LONG)height*(LONG)lpDBTTip->uMe.nHotSpotY) / 100L  );
  SelectObject( hDC, CreateSolidBrush( RGB(255,0,0) ) );
  Ellipse( hDC, nHotSpotX-2, nHotSpotY-2, nHotSpotX+3, nHotSpotY+3 );
  DeleteObject( SelectObject( hDC, GetStockObject(NULL_BRUSH) ) );
  DrawIcon( hDC,
            nHotSpotX-lpDBTTip->uMe.nMouseOffX,
            nHotSpotY-lpDBTTip->uMe.nMouseOffY,
            LoadIcon( ghmodDLL, MAKEINTRESOURCE(IDI_ARROW) ) );
 }

 //The bounding box of the text
 rect.right	= rect.left + (short)( ((LONG)width*(LONG)lpDBTTip->uMe.nTextRight) / 100L );
 rect.bottom= rect.top + (short)( ((LONG)height*(LONG)lpDBTTip->uMe.nTextBottom) / 100L );
 rect.left	+= (short)( ((LONG)width*(LONG)lpDBTTip->uMe.nTextLeft) / 100L  );
 rect.top	+= (short)( ((LONG)height*(LONG)lpDBTTip->uMe.nTextTop) / 100L );

 hOldPen = SelectObject( hDC, GetStockObject(BLACK_PEN) );
 Rectangle( hDC, rect.left, rect.top, rect.right, rect.bottom );
 hOldPen = SelectObject( hDC, GetStockObject(WHITE_PEN) );
 Rectangle( hDC, rect.left+1, rect.top+1, rect.right-1, rect.bottom-1 );

 //Restore to normal (whatever that may be!!!!)
 SelectObject( hDC, hOldPen );
 SelectObject( hDC, hOldBrush );
}


BOOL DoGraphicalHitTest( POINT p, HCTL hctl, LPRECT lprectCapture )
{
 HITTEST HitTest;
 POINT ptParentUL;
 HCTL hctlParent;

 // copy this rect in case the lpDBTTip struct will move
 HitTest.rect = *lprectCapture;

 if ( VBGetControlHwnd(hctl) )  // is it a graphical control?
   return TRUE;  // no - then the test can't fail

 hctlParent = VBGetControl( hctl, GC_CONTAINER );
 if ( !hctlParent )
   return TRUE;  // can't continue if no parent (must be an error...)

 ptParentUL.x = 0;
 ptParentUL.y = 0;
 VBClientToScreen( hctlParent, &ptParentUL );
 OffsetRect( &HitTest.rect, -ptParentUL.x, -ptParentUL.y );

 HitTest.pt = p;
 VBScreenToClient( hctlParent, &HitTest.pt );

 return VBSendControlMsg( hctl, VBM_HITTEST, 0, (LONG)(LPVOID)&HitTest ) != HT_MISS ;
}


HCTL SearchContainer( HCTL hctl, POINT pM)
{
 RECT rect;

 if ( !hctl )
   return NULL;
 hctl = VBGetControl( hctl, GC_CHILD );
 while ( hctl )
  {
   VBGetControlRect( hctl, &rect );
   if ( PtInRect(&rect,pM) && VBIsControlVisible(hctl) )
     return hctl;
   hctl = VBGetControl( hctl, GC_NEXTSIBLING );
  }
 return NULL;
}


HCTL FindCorrectSibling( HCTL hctl, short nIndex )
{
 MODEL FAR *lpModel;
 PPROPINFO FAR *lpPropInfo;
 char szN1[41], szN2[41];
 LONG lIndex;
 short i;

 if ( nIndex==0 )
   return hctl;

 VBGetControlName( hctl, szN1 );
 hctl = VBGetControl( hctl, GC_FIRSTSIBLING );
 while ( hctl )
  {
   VBGetControlName( hctl, szN2 );
   if ( (lpModel=VBGetControlModel(hctl)) && !lstrcmp(szN1,szN2) )
    {
     lpPropInfo = (LPVOID)MAKELONG( lpModel->npproplist, (_segment)lpModel );
     for ( i=0 ; *lpPropInfo ; lpPropInfo++, i++ )
       if ( *lpPropInfo == PPROPINFO_STD_INDEX )
        {
         if ( !VBGetControlProperty( hctl, i, &lIndex ) )
           if ( nIndex == (short)lIndex )
             return hctl;
         break;
        }
    }
   hctl = VBGetControl( hctl, GC_NEXTSIBLING );
  }

 return NULL;
}


HLSTR SuggestTipText( HCTL hctl, short FAR *lpnIndex, short FAR *lpnUseIndex )
{
 MODEL FAR *lpModel;
 PPROPINFO FAR *lpPropInfo;
 short nTagIndex=-1, nIndexIndex=-1, i;

 *lpnUseIndex = 0;
 lpModel = VBGetControlModel( hctl );
 if ( !lpModel )
   return VBCreateHlstr( NULL, 0 );
 lpPropInfo = (LPVOID)MAKELONG( lpModel->npproplist, (_segment)lpModel );

 for ( i=0 ; *lpPropInfo ; lpPropInfo++, i++ )
  {
   if ( *lpPropInfo == PPROPINFO_STD_INDEX )
     nIndexIndex = i;
   if ( *lpPropInfo == PPROPINFO_STD_TAG )
     nTagIndex = i;
  }

 *lpnIndex = 0;
 if ( nIndexIndex>=0 )
  {
   LONG lIndex;
   if ( !VBGetControlProperty( hctl, nIndexIndex, &lIndex ) )
     *lpnIndex = (short)lIndex;
  }

 if ( nTagIndex>=0 )
  {
   HSZ hsz;
   if ( !VBGetControlProperty( hctl, nTagIndex, &hsz ) )
    {
     LPSTR lpszT = VBDerefHsz(hsz);
     char szHead[10];
     lstrcpyn( szHead, lpszT, 7 );
     if ( !lstrcmp( szHead, szDBTTip ) )
      {
       for ( i=6 ; ; i++ )
         switch ( lpszT[i] )
          {
           case ':': {
             HLSTR hlStr = VBCreateHlstr( lpszT+i+1, lstrlen(lpszT)-i-1 );
             VBDestroyHsz(hsz);
             return hlStr; }
           case '0': case '1': case '2': case '3': case '4':
           case '5': case '6': case '7': case '8': case '9':
             *lpnUseIndex = *lpnUseIndex*10 + lpszT[i]-'0';
             break;
           default:
             VBDestroyHsz(hsz);
             return VBCreateHlstr( NULL, 0 );
          }
      }
     VBDestroyHsz(hsz);
    }
  }

 return VBCreateHlstr( NULL, 0 );
}


void StandingStill(HCTL hctl, POINT pM)
{
 LPDBTTIP lpDBTTip;
 HCTL hHit;
 RECT rectCapture;

 // Verify that the window the mouse is parked over is a VB control
 {
  HWND hwndMouseOver = WindowFromPoint( pM );
  if ( !hwndMouseOver )
    return;
  hHit = VBGetHwndControl( hwndMouseOver );
  if ( !hHit )
   {  // Maybe it is the edit control of a combobox... check out it's parent also...
    hwndMouseOver = GetWindowWord( hwndMouseOver, GWW_HWNDPARENT );
    if ( !hwndMouseOver )
      return;
    hHit = VBGetHwndControl( hwndMouseOver );
      if ( !hHit )
        return;
   }
 }

 // We now know the hctl of the VB control - or it's container,
 // if it's a graphical control!

 // Verify that the newly found control belongs to the same
 // application instance as we do and that it's form isn't
 // minimized
 {
  HWND hwndF = VBGetControlHwnd( VBGetControl( hHit, GC_FORM) );
  if ( GetWindowWord( hwndF, GWW_HINSTANCE ) != VBGetHInstance() )
    return;
  if ( IsIconic( hwndF ) )  // v1.23 950628
    return;
  if ( !IsWindowEnabled( hwndF ) )  // v1.31 951007
    return;
 }

 // Now we narrow the search 'til we know that it's not the container
 // of the desired control we've found ( WindowFromPoint knows nothing
 // of VB graphical controls )
 {
  HCTL hDeeper;
  while ( hDeeper = SearchContainer( hHit, pM ) )
    hHit = hDeeper;
 }

 // We don't want a form to get tipped
 if ( hHit == VBGetControl( hHit, GC_FORM ) )
   return;

 // Now get the control's _client_ area in screen coordinates.
 // Note: Using is's _window_ area will yield wrong results
 // when we send WM_MOUSEMOVE to it later.
 {
  POINT ptClientUL;
  ptClientUL.x = 0;
  ptClientUL.y = 0;
  VBClientToScreen( hHit, &ptClientUL );
  VBGetClientRect( hHit, &rectCapture );
  OffsetRect( &rectCapture, ptClientUL.x, ptClientUL.y );
  if ( !PtInRect(&rectCapture,pM ) || !DoGraphicalHitTest(pM,hHit,&rectCapture) )
    return;
 }

 {
  PARAMS p;
  LPSTR lpszT;
  HWND hWnd = VBGetControlHwnd( hHit );
  float xTwips = (float)VBXPixelsToTwips( pM.x - rectCapture.left );
  float yTwips = (float)VBYPixelsToTwips( pM.y - rectCapture.top );
  char szCtlName[41];
  short nIndex, nUseIndex;

  VBGetControlName( hHit, szCtlName );
  p.hlCtlName = VBCreateHlstr( szCtlName, lstrlen(szCtlName) );
  p.lphWnd = &hWnd;
  p.hlStr = SuggestTipText( hHit, &nIndex, &nUseIndex );
  LPDBTTIPDEREF(hctl)->nUseIndex = nUseIndex;
  LPDBTTIPDEREF(hctl)->bCtlEnabled = !!VBIsControlEnabled( hHit );  //950916 v1.3
  p.lpnIndex = &nIndex;
  p.x = &xTwips;
  p.y = &yTwips;

  if ( VBFireEvent( hctl, IEVENT_TTIP_POPUP, &p ) )
   {
    VBDestroyHlstr( p.hlStr );
    VBDestroyHlstr( p.hlCtlName );
    return;
   }

  VBDestroyHlstr( p.hlCtlName );
  if ( LPDBTTIPDEREF(hctl)->hszText )                //v1.24
   {                                                 //
    VBDestroyHsz( LPDBTTIPDEREF(hctl)->hszText );    //
    LPDBTTIPDEREF(hctl)->hszText = NULL;             //
   }                                                 //
  lpszT = VBDerefZeroTermHlstr( p.hlStr );
  if ( !lpszT )
    return;
  if ( !*lpszT )
   {
    VBDestroyHlstr( p.hlStr );
    return;
   }
  //lstrcpyn( LPDBTTIPDEREF(hctl)->szText, lpszT, 255 );  v1.24
  LPDBTTIPDEREF(hctl)->hszText = VBCreateHsz( (_segment)hctl, lpszT );  //v1.24
  VBDestroyHlstr( p.hlStr );
 }

 // now store the hwnd of a known container (or the control or the form)
 {
  HCTL hctlContainer = hHit;
  HWND hwnd;
  while ( !(hwnd=VBGetControlHwnd(hctlContainer)) )
   {
    hctlContainer = VBGetControl( hctlContainer, GC_CONTAINER );
    if ( !hctlContainer )
      hctlContainer = VBGetControl( hHit, GC_FORM );
   }
  LPDBTTIPDEREF(hctl)->hwndKnownContainer = hwnd ;
 }

 // now move the correct "uMe" to "uUse"
 {
  HCTL hctlUse = FindCorrectSibling( hctl, LPDBTTIPDEREF(hctl)->nUseIndex );
  if ( hctlUse )
   {
    UNIQUE u = LPDBTTIPDEREF(hctlUse)->uMe;
    lpDBTTip = LPDBTTIPDEREF(hctl);
    lpDBTTip->uUse = u;
    lpDBTTip->TipStatus = TIPPING_VISIBLE;
   }
  else
   {
    lpDBTTip = LPDBTTIPDEREF(hctl);
    lpDBTTip->uUse = lpDBTTip->uMe;		// "default" settings
    lpDBTTip->uUse.nTractorBeam = 0;		// no beaming for invisible tips
    lpDBTTip->TipStatus = TIPPING_INVISIBLE;
   }
 }

 lpDBTTip->hctlTipping = hHit;
 lpDBTTip->pMouse = pM;
 lpDBTTip->pMouseOrgPos = pM;
 CopyRect( &lpDBTTip->rectCapture, &rectCapture );

 if ( LPDBTTIPDEREF(hctl)->TipStatus == TIPPING_VISIBLE )
  {
   AdjustTipSize( hctl );
   CreateTipWnd( hctl );
  }
 SetCapture( VBGetControlHwnd(hctl) );
}


//---------------------------------------------------------------------------
// TTip Control Procedure
//---------------------------------------------------------------------------
LONG FAR PASCAL _export TTipCtlProc
(
 HCTL   hctl,
 HWND   hWnd,
 USHORT msg,
 USHORT wp,
 LONG   lp
)
{
 LPDBTTIP lpDBTTip = LPDBTTIPDEREF(hctl);

 // Message pre-processing
 switch ( msg )
  {
   case WM_NCCREATE:
     lpDBTTip->uMe.nTextLeft	= 5;
     lpDBTTip->uMe.nTextTop		= 5;
     lpDBTTip->uMe.nTextRight	= 95;
     lpDBTTip->uMe.nTextBottom	= 75;
     lpDBTTip->uMe.nHotSpotX	= 100;
     lpDBTTip->uMe.nHotSpotY	= 100;
     lpDBTTip->uMe.nMouseOffX	= -3;
     lpDBTTip->uMe.nMouseOffY	= -3;
     lpDBTTip->uMe.nAspect		= 20;
     lpDBTTip->uMe.nAlignment	= 2;
     lpDBTTip->nCountDown = lpDBTTip->nDelay / 50;
	 VBSetControlProperty(hctl, IPROP_TTIP_ABOUTBOX, (LONG)(LPSTR)"Click on \"...\" for the About Box ---->");
     break;


   case WM_NCDESTROY:
     KillTimer( hWnd, 1 );
     if ( lpDBTTip->hWndTip )
       DestroyWindow( lpDBTTip->hWndTip );
     if ( lpDBTTip->uMe.hPic )
      {
       VBFreePic(lpDBTTip->uMe.hPic);
       lpDBTTip = LPDBTTIPDEREF(hctl);
      }
     if ( lpDBTTip->hszAbout )
      {
       VBDestroyHsz(lpDBTTip->hszAbout);
       lpDBTTip = LPDBTTIPDEREF(hctl);
      }
     if ( lpDBTTip->hszAbout )
       VBDestroyHsz(lpDBTTip->hszText);
     break;


   case VBM_CREATED:
	 if ( VBGetMode() != MODE_DESIGN )
       return 0L;
     break;


   case WM_SETFONT:
	 lpDBTTip->uMe.hFont = (HFONT)wp;
     return 0;


   case WM_GETFONT:
	 return lpDBTTip->uMe.hFont;


   case VBM_GETPROPERTY:
     switch ( wp )
      {
	   case IPROP_TTIP_ISVISIBLE:
	     *(BOOL FAR *)lp = lpDBTTip->TipStatus != TIPPING_NONE;
	     return 0;
	  }
     break;

   case VBM_SETPROPERTY:
     switch ( wp )
      {
	   case IPROP_TTIP_ASPECT:
	     if ( (short)lp < 1 )
	       return ERR_InvPropVal;
	     return 0L;
	   case IPROP_TTIP_TEXTLEFT:
	   case IPROP_TTIP_TEXTTOP:
	   case IPROP_TTIP_TEXTRIGHT:
	   case IPROP_TTIP_TEXTBOTTOM:
	   case IPROP_TTIP_HOTSPOTX:
	   case IPROP_TTIP_HOTSPOTY:
	   case IPROP_TTIP_MOUSEOFFX:
	   case IPROP_TTIP_MOUSEOFFY:
	     if ( VBGetMode() == MODE_DESIGN )
           InvalidateRect(hWnd, NULL, TRUE);
         return 0L;

	   case IPROP_TTIP_POPUPMODE:
         if ( VBGetMode() == MODE_RUN )
           if( (short)lp == PMODE_MANUAL )
             KillTimer( hWnd, 1 );
           else
             SetTimer( hWnd, 1, 50, NULL );
         return 0L;

       case  IPROP_TTIP_ISVISIBLE:
         if ( GetCapture()==hWnd )
          {
           if ( lpDBTTip->PopupMode == PMODE_MANUAL )
             KillTimer( hWnd, 1 );
           ReleaseCapture();
           if ( GoDown(hctl) )
             return 0L;
          }
         if ( (BOOL)lp )
          {
           if ( lpDBTTip->PopupMode == PMODE_MANUAL )
             SetTimer( hWnd, 1, 50, NULL );
           lpDBTTip->nCountDown = FAST_POPUP;
          }
         return 0L;
      }
     break;

   case VBM_CHECKPROPERTY:
     switch ( wp )
      {
	   case IPROP_TTIP_PICTURE: {
	     PIC pic;

         VBGetPic( (HPIC)lp, &pic);
		 switch ( pic.picType )
		  {
		   case PICTYPE_NONE:
		   case PICTYPE_METAFILE:
             InvalidateRect( hWnd, NULL, TRUE );
		     return 0;
          }
         return ERR_InvPropVal; }
      }
	 break;


   case VBM_INITPROPPOPUP:
	 if ( wp == IPROP_TTIP_ABOUTBOX )
	  {
	   HWND hwndList = (HWND)LOWORD(lp);
	   return (LONG)HwndInitAboutBox();
      }
     break;

   case WM_PAINT: {
     RECT rect;
     PIC pic;
     GetClientRect( hWnd, &rect );
     VBGetPic( lpDBTTip->uMe.hPic, &pic );
     if ( pic.picType!=PICTYPE_METAFILE )
       break;
     if  ( wp )
       PaintTip( hctl, (HDC)wp, rect, pic.picData.wmf.hmeta, pic.picData.wmf.xExt, pic.picData.wmf.yExt );
     else
      {
       PAINTSTRUCT ps;
       BeginPaint( hWnd, &ps );
       PaintTip( hctl, ps.hdc, rect, pic.picData.wmf.hmeta, pic.picData.wmf.xExt, pic.picData.wmf.yExt );
       EndPaint( hWnd, &ps );
      }
     return 0; }


   case WM_TIMER: {
     POINT p;
     short dx, dy;
     GetCursorPos(&p);

     if ( lpDBTTip->TipStatus != TIPPING_NONE )  // Tip is up
      {
       if ( GetKeyState(VK_SHIFT)<0 || GetKeyState(VK_MENU)<0 || GetKeyState(VK_CONTROL)<0 )
         ReleaseCapture();
       if ( GetCapture() != hWnd )  // Have we, for some reason, lost the capture?
         GoDown( hctl );     		// then we must go away
       return 0L;
      }

     if ( VBGetMode()!=MODE_RUN )  // v1.23 950627
       return 0L;

     if ( lpDBTTip->nCountDown==FAST_POPUP )
      {
       StandingStill( hctl, p );
       lpDBTTip->nCountDown = -1;
       return 0L;
      }

     if ( lpDBTTip->PopupMode==PMODE_MANUAL )
      {
       KillTimer( hWnd, 1 );
       return 0L;
      }

     dx = p.x - lpDBTTip->pMouse.x;
     dy = p.y - lpDBTTip->pMouse.y;
     lpDBTTip->pMouse = p;
     if ( (dx*dx + dy*dy) <= (lpDBTTip->uUse.nSensivity*lpDBTTip->uUse.nSensivity) && lpDBTTip->nCountDown>=0 )
      {
       if ( GetKeyState(VK_LBUTTON)>=0 && GetKeyState(VK_MBUTTON)>=0 && GetKeyState(VK_RBUTTON)>=0 )
         if ( !GetCapture() )  // don't popup if the mouse is already captured by someone
           if ( !lpDBTTip->nCountDown-- )
             StandingStill( hctl, p );
      }
     else
       lpDBTTip->nCountDown = lpDBTTip->nDelay / 50;
     return 0; }


   case WM_MOUSEMOVE:
     if ( lpDBTTip->TipStatus != TIPPING_NONE )  // make sure there really is a Tip up (error otherwise...)
      {
       POINT p;
       BOOL bMiss;

       if ( !VerifyHCTL( lpDBTTip->hctlTipping, lpDBTTip->hwndKnownContainer ) )
        {
         ReleaseCapture();
         GoDown( hctl );
         return 0L;
        }

       GetCursorPos(&p);
       lpDBTTip = LPDBTTIPDEREF(hctl);
       bMiss = !DoGraphicalHitTest( p, lpDBTTip->hctlTipping, &lpDBTTip->rectCapture );
       lpDBTTip = LPDBTTIPDEREF( hctl );
       if ( !PtInRect( &lpDBTTip->rectCapture, p ) ||  bMiss )  // fallen outside the rect?
        { UpNDown:  // some rearrangement 1.32
         ReleaseCapture();
         GoDown( hctl );
         if ( lpDBTTip->PopupMode == PMODE_AUTOMATIC )
           lpDBTTip->nCountDown = FAST_POPUP;
         return 0L;
        }

       if ( SearchContainer( lpDBTTip->hctlTipping, p ) )  // moved into a child control?
         if ( lpDBTTip->PopupMode == PMODE_AUTOMATIC )
           goto UpNDown;
       lpDBTTip = LPDBTTIPDEREF(hctl);
       if ( lpDBTTip->uUse.nTractorBeam )  // beaming turned on?
        {
         short dx = p.x - lpDBTTip->pMouseOrgPos.x;
         short dy = p.y - lpDBTTip->pMouseOrgPos.y;
         if ( (dx*dx + dy*dy) >= lpDBTTip->uUse.nTractorBeam*lpDBTTip->uUse.nTractorBeam )
           goto UpNDown;
        }
       if ( VBIsControlEnabled(lpDBTTip->hctlTipping) )
        {
         VBScreenToClient( LPDBTTIPDEREF(hctl)->hctlTipping, &p );
         VBSendControlMsg( LPDBTTIPDEREF(hctl)->hctlTipping, msg, wp, MAKELONG(p.x,p.y) );
        }
      }
     return 0;

   case WM_LBUTTONDOWN:
   case WM_MBUTTONDOWN:
   case WM_RBUTTONDOWN:
     if ( lpDBTTip->TipStatus != TIPPING_NONE )
      {
       POINT p;
       GetCursorPos(&p);
	   ReleaseCapture();
       if ( !GoDown(hctl) )
         if ( VerifyHCTL( lpDBTTip->hctlTipping, lpDBTTip->hwndKnownContainer ) )
           if ( VBIsControlEnabled( LPDBTTIPDEREF(hctl)->hctlTipping ) )
            {
             VBScreenToClient( LPDBTTIPDEREF(hctl)->hctlTipping, &p );
             VBSendControlMsg( LPDBTTIPDEREF(hctl)->hctlTipping, msg, wp, MAKELONG(p.x,p.y) );
            }
      }
     return 0;

//   case 0x105C:
//     return 0;
  }

 // Default processing:
 return VBDefControlProc(hctl, hWnd, msg, wp, lp);
}


////////////////////////////////////////////////////////////


WORD FAR PASCAL __export DBTTipVersion()
{
 return VBX_VERSION_WORD;
}


//---------------------------------------------------------------------------
// Register custom control. This routine is called by VB when the custom
// control DLL is loaded for use.
//---------------------------------------------------------------------------
BOOL FAR PASCAL _export VBINITCC
(
 USHORT usVersion,
 BOOL   fRuntime
)
{
 extern MODEL modelTTip;
 char szStr[300], FAR *ptr;
 DWORD dwT, dwC=0;

 LoadString( ghmodDLL, IDS_WHOBUILTME, szStr, sizeof(szStr)-1 );
 if ( !fRuntime && ((GetTickCount()/1000)&7)==0 )  // Tell the world about the author
  {
   do {  // If the MessageBox executes too fast, maybe someone has patched the code?
     dwT=GetTickCount()+500;
     MessageBox(GetFocus(),szStr,"DBTTip v1.32 Nov 1995",MB_OK|MB_ICONEXCLAMATION);
    } while (  dwT>GetTickCount() );
  }
 else
  {
   for ( ptr=&szStr[0] ; *ptr ; ptr++ )
    {
     if ( (long)dwC < 0 )
       dwC++;
     dwC = (dwC<<1)+(*ptr);
    }
   if ( dwC != 3340965503 )
     return FALSE;
  }

 if ( !fRuntime )
  {
   WNDCLASS class;
   class.style          = 0;
   class.lpfnWndProc    = AboutBoxProc;
   class.cbClsExtra     = 0;
   class.cbWndExtra     = 0;
   class.hInstance      = ghmodDLL;
   class.hIcon          = LoadIcon(ghmodDLL,MAKEINTRESOURCE(IDI_DBTTIP));
   class.hCursor        = NULL;
   class.hbrBackground  = NULL;
   class.lpszMenuName   = NULL;
   class.lpszClassName  = CLASS_ABOUTBOX;
   RegisterClass(&class);
   taskDevEnvironment = GetCurrentTask();
  }

 if ( !RegisterTipClass() )
   return FALSE;

 return VBRegisterModel( ghmodDLL, &modelTTip );
}



void FAR PASCAL _export VBTERMCC()
{
 if ( taskDevEnvironment == GetCurrentTask() )
  {
   UnregisterClass(CLASS_ABOUTBOX, ghmodDLL);
   taskDevEnvironment = NULL;
  }

 UnregisterTipClass();
}


//---------------------------------------------------------------------------
// Provide custom control model information to host environment.
//---------------------------------------------------------------------------
LPMODELINFO FAR PASCAL _export VBGetModelInfo
(
 USHORT usVersion
)
{
 extern MODELINFO modelinfoTTip;
 return &modelinfoTTip;
}


//---------------------------------------------------------------------------
// Initialize library. This routine is called when the first client loads
// the DLL.
//---------------------------------------------------------------------------
int FAR PASCAL LibMain
(
 HANDLE hModule,
 WORD   wDataSeg,
 WORD   cbHeapSize,
 LPSTR  lpszCmdLine
)
{
 // Avoid warnings on unused (but required) formal parameters
 wDataSeg = wDataSeg;
 cbHeapSize = cbHeapSize;
 lpszCmdLine = lpszCmdLine;

 ghmodDLL = hModule;

 return 1;
}
